# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import argparse
import sys
from difflib import get_close_matches

AWS_CLI_V2_MESSAGE = (
    'Note: AWS CLI version 2, the latest major version '
    'of the AWS CLI, is now stable and recommended for general '
    'use. For more information, see the AWS CLI version 2 '
    'installation instructions at: https://docs.aws.amazon.com/cli/'
    'latest/userguide/install-cliv2.html'
)

HELP_BLURB = (
    "To see help text, you can run:\n"
    "\n"
    "  aws help\n"
    "  aws <command> help\n"
    "  aws <command> <subcommand> help\n"
)
USAGE = (
    "\r%s\n\n"
    "usage: aws [options] <command> <subcommand> "
    "[<subcommand> ...] [parameters]\n"
    "%s" % (AWS_CLI_V2_MESSAGE, HELP_BLURB)
)


class CommandAction(argparse.Action):
    """Custom action for CLI command arguments

    Allows the choices for the argument to be mutable. The choices
    are dynamically retrieved from the keys of the referenced command
    table
    """

    def __init__(self, option_strings, dest, command_table, **kwargs):
        self.command_table = command_table
        super().__init__(
            option_strings, dest, choices=self.choices, **kwargs
        )

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)

    @property
    def choices(self):
        return list(self.command_table.keys())

    @choices.setter
    def choices(self, val):
        # argparse.Action will always try to set this value upon
        # instantiation, but this value should be dynamically
        # generated from the command table keys. So make this a
        # NOOP if argparse.Action tries to set this value.
        pass


class CLIArgParser(argparse.ArgumentParser):
    Formatter = argparse.RawTextHelpFormatter

    # When displaying invalid choice error messages,
    # this controls how many options to show per line.
    ChoicesPerLine = 2

    def _check_value(self, action, value):
        """
        It's probably not a great idea to override a "hidden" method
        but the default behavior is pretty ugly and there doesn't
        seem to be any other way to change it.
        """
        # converted value must be one of the choices (if specified)
        if action.choices is not None and value not in action.choices:
            msg = ['Invalid choice, valid choices are:\n']
            for i in range(len(action.choices))[:: self.ChoicesPerLine]:
                current = []
                for choice in action.choices[i : i + self.ChoicesPerLine]:
                    current.append('%-40s' % choice)
                msg.append(' | '.join(current))
            possible = get_close_matches(value, action.choices, cutoff=0.8)
            if possible:
                extra = ['\n\nInvalid choice: %r, maybe you meant:\n' % value]
                for word in possible:
                    extra.append('  * %s' % word)
                msg.extend(extra)
            raise argparse.ArgumentError(action, '\n'.join(msg))

    def parse_known_args(self, args, namespace=None):
        parsed, remaining = super().parse_known_args(
            args, namespace
        )
        terminal_encoding = getattr(sys.stdin, 'encoding', 'utf-8')
        if terminal_encoding is None:
            # In some cases, sys.stdin won't have an encoding set,
            # (e.g if it's set to a StringIO).  In this case we just
            # default to utf-8.
            terminal_encoding = 'utf-8'
        for arg, value in vars(parsed).items():
            if isinstance(value, bytes):
                setattr(parsed, arg, value.decode(terminal_encoding))
            elif isinstance(value, list):
                encoded = []
                for v in value:
                    if isinstance(v, bytes):
                        encoded.append(v.decode(terminal_encoding))
                    else:
                        encoded.append(v)
                setattr(parsed, arg, encoded)
        return parsed, remaining


class MainArgParser(CLIArgParser):
    Formatter = argparse.RawTextHelpFormatter

    def __init__(
        self,
        command_table,
        version_string,
        description,
        argument_table,
        prog=None,
    ):
        super().__init__(
            formatter_class=self.Formatter,
            add_help=False,
            conflict_handler='resolve',
            description=description,
            usage=USAGE,
            prog=prog,
        )
        self._build(command_table, version_string, argument_table)

    def _create_choice_help(self, choices):
        help_str = ''
        for choice in sorted(choices):
            help_str += '* %s\n' % choice
        return help_str

    def _build(self, command_table, version_string, argument_table):
        for argument_name in argument_table:
            argument = argument_table[argument_name]
            argument.add_to_parser(self)
        self.add_argument(
            '--version',
            action="version",
            version=version_string,
            help='Display the version of this tool',
        )
        self.add_argument(
            'command', action=CommandAction, command_table=command_table
        )


class ServiceArgParser(CLIArgParser):
    def __init__(self, operations_table, service_name):
        super().__init__(
            formatter_class=argparse.RawTextHelpFormatter,
            add_help=False,
            conflict_handler='resolve',
            usage=USAGE,
        )
        self._build(operations_table)
        self._service_name = service_name

    def _build(self, operations_table):
        self.add_argument(
            'operation', action=CommandAction, command_table=operations_table
        )


class ArgTableArgParser(CLIArgParser):
    """CLI arg parser based on an argument table."""

    def __init__(self, argument_table, command_table=None):
        # command_table is an optional subcommand_table.  If it's passed
        # in, then we'll update the argparse to parse a 'subcommand' argument
        # and populate the choices field with the command table keys.
        super().__init__(
            formatter_class=self.Formatter,
            add_help=False,
            usage=USAGE,
            conflict_handler='resolve',
        )
        if command_table is None:
            command_table = {}
        self._build(argument_table, command_table)

    def _build(self, argument_table, command_table):
        for arg_name in argument_table:
            argument = argument_table[arg_name]
            argument.add_to_parser(self)
        if command_table:
            self.add_argument(
                'subcommand',
                action=CommandAction,
                command_table=command_table,
                nargs='?',
            )

    def parse_known_args(self, args, namespace=None):
        if len(args) == 1 and args[0] == 'help':
            namespace = argparse.Namespace()
            namespace.help = 'help'
            return namespace, []
        else:
            return super().parse_known_args(
                args, namespace
            )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class ECSError(Exception):
    """ Base class for all ECSErrors."""
    fmt = 'An unspecified error occurred'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        super(ECSError, self).__init__(msg)
        self.kwargs = kwargs


class MissingPropertyError(ECSError):
    fmt = \
        "Error: Resource '{resource}' must include property '{prop_name}'"


class FileLoadError(ECSError):
    fmt = "Error: Unable to load file at {file_path}: {error}"


class InvalidPlatformError(ECSError):
    fmt = "Error: {resource} '{name}' must support 'ECS' compute platform"


class InvalidProperyError(ECSError):
    fmt = ("Error: deployment group '{dg_name}' does not target "
           "ECS {resource} '{resource_name}'")


class InvalidServiceError(ECSError):
    fmt = "Error: Service '{service}' not found in cluster '{cluster}'"


class ServiceClientError(ECSError):
    fmt = "Failed to {action}:\n{error}"


# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os
import platform
import shlex
import sys
from subprocess import PIPE, Popen

from docutils.core import publish_string
from docutils.writers import manpage

from awscli.argparser import ArgTableArgParser
from awscli.argprocess import ParamShorthandParser
from awscli.bcdoc import docevents
from awscli.bcdoc.restdoc import ReSTDocument
from awscli.bcdoc.textwriter import TextWriter
from awscli.clidocs import (
    OperationDocumentEventHandler,
    ProviderDocumentEventHandler,
    ServiceDocumentEventHandler,
    TopicDocumentEventHandler,
    TopicListerDocumentEventHandler,
)
from awscli.topictags import TopicTagDB
from awscli.utils import ignore_ctrl_c

LOG = logging.getLogger('awscli.help')


class ExecutableNotFoundError(Exception):
    def __init__(self, executable_name):
        super().__init__(
            f'Could not find executable named "{executable_name}"'
        )


def get_renderer():
    """
    Return the appropriate HelpRenderer implementation for the
    current platform.
    """
    if platform.system() == 'Windows':
        return WindowsHelpRenderer()
    else:
        return PosixHelpRenderer()


class PagingHelpRenderer:
    """
    Interface for a help renderer.

    The renderer is responsible for displaying the help content on
    a particular platform.

    """

    def __init__(self, output_stream=sys.stdout):
        self.output_stream = output_stream

    PAGER = None
    _DEFAULT_DOCUTILS_SETTINGS_OVERRIDES = {
        # The default for line length limit in docutils is 10,000. However,
        # currently in the documentation, it inlines all possible enums in
        # the JSON syntax which exceeds this limit for some EC2 commands
        # and prevents the manpages from being generated.
        # This is a temporary fix to allow the manpages for these commands
        # to be rendered. Long term, we should avoid enumerating over all
        # enums inline for the JSON syntax snippets.
        'line_length_limit': 50_000
    }

    def get_pager_cmdline(self):
        pager = self.PAGER
        if 'MANPAGER' in os.environ:
            pager = os.environ['MANPAGER']
        elif 'PAGER' in os.environ:
            pager = os.environ['PAGER']
        return shlex.split(pager)

    def render(self, contents):
        """
        Each implementation of HelpRenderer must implement this
        render method.
        """
        converted_content = self._convert_doc_content(contents)
        self._send_output_to_pager(converted_content)

    def _send_output_to_pager(self, output):
        cmdline = self.get_pager_cmdline()
        LOG.debug("Running command: %s", cmdline)
        p = self._popen(cmdline, stdin=PIPE)
        p.communicate(input=output)

    def _popen(self, *args, **kwargs):
        return Popen(*args, **kwargs)

    def _convert_doc_content(self, contents):
        return contents


class PosixHelpRenderer(PagingHelpRenderer):
    """
    Render help content on a Posix-like system.  This includes
    Linux and MacOS X.
    """

    PAGER = 'less -R'

    def _convert_doc_content(self, contents):
        man_contents = publish_string(
            contents,
            writer=manpage.Writer(),
            settings_overrides=self._DEFAULT_DOCUTILS_SETTINGS_OVERRIDES,
        )
        if self._exists_on_path('groff'):
            cmdline = ['groff', '-m', 'man', '-T', 'ascii']
        elif self._exists_on_path('mandoc'):
            cmdline = ['mandoc', '-T', 'ascii']
        else:
            raise ExecutableNotFoundError('groff or mandoc')
        LOG.debug("Running command: %s", cmdline)
        p3 = self._popen(cmdline, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output = p3.communicate(input=man_contents)[0]
        return output

    def _send_output_to_pager(self, output):
        cmdline = self.get_pager_cmdline()
        if not self._exists_on_path(cmdline[0]):
            LOG.debug(
                "Pager '%s' not found in PATH, printing raw help.", cmdline[0]
            )
            self.output_stream.write(output.decode('utf-8') + "\n")
            self.output_stream.flush()
            return
        LOG.debug("Running command: %s", cmdline)
        with ignore_ctrl_c():
            # We can't rely on the KeyboardInterrupt from
            # the CLIDriver being caught because when we
            # send the output to a pager it will use various
            # control characters that need to be cleaned
            # up gracefully.  Otherwise if we simply catch
            # the Ctrl-C and exit, it will likely leave the
            # users terminals in a bad state and they'll need
            # to manually run ``reset`` to fix this issue.
            # Ignoring Ctrl-C solves this issue.  It's also
            # the default behavior of less (you can't ctrl-c
            # out of a manpage).
            p = self._popen(cmdline, stdin=PIPE)
            p.communicate(input=output)

    def _exists_on_path(self, name):
        # Since we're only dealing with POSIX systems, we can
        # ignore things like PATHEXT.
        return any(
            [
                os.path.exists(os.path.join(p, name))
                for p in os.environ.get('PATH', '').split(os.pathsep)
            ]
        )


class WindowsHelpRenderer(PagingHelpRenderer):
    """Render help content on a Windows platform."""

    PAGER = 'more'

    def _convert_doc_content(self, contents):
        text_output = publish_string(
            contents,
            writer=TextWriter(),
            settings_overrides=self._DEFAULT_DOCUTILS_SETTINGS_OVERRIDES,
        )
        return text_output

    def _popen(self, *args, **kwargs):
        # Also set the shell value to True.  To get any of the
        # piping to a pager to work, we need to use shell=True.
        kwargs['shell'] = True
        return Popen(*args, **kwargs)


class HelpCommand:
    """
    HelpCommand Interface
    ---------------------
    A HelpCommand object acts as the interface between objects in the
    CLI (e.g. Providers, Services, Operations, etc.) and the documentation
    system (bcdoc).

    A HelpCommand object wraps the object from the CLI space and provides
    a consistent interface to critical information needed by the
    documentation pipeline such as the object's name, description, etc.

    The HelpCommand object is passed to the component of the
    documentation pipeline that fires documentation events.  It is
    then passed on to each document event handler that has registered
    for the events.

    All HelpCommand objects contain the following attributes:

        + ``session`` - A ``botocore`` ``Session`` object.
        + ``obj`` - The object that is being documented.
        + ``command_table`` - A dict mapping command names to
              callable objects.
        + ``arg_table`` - A dict mapping argument names to callable objects.
        + ``doc`` - A ``Document`` object that is used to collect the
              generated documentation.

    In addition, please note the `properties` defined below which are
    required to allow the object to be used in the document pipeline.

    Implementations of HelpCommand are provided here for Provider,
    Service and Operation objects.  Other implementations for other
    types of objects might be needed for customization in plugins.
    As long as the implementations conform to this basic interface
    it should be possible to pass them to the documentation system
    and generate interactive and static help files.
    """

    EventHandlerClass = None
    """
    Each subclass should define this class variable to point to the
    EventHandler class used by this HelpCommand.
    """

    def __init__(self, session, obj, command_table, arg_table):
        self.session = session
        self.obj = obj
        if command_table is None:
            command_table = {}
        self.command_table = command_table
        if arg_table is None:
            arg_table = {}
        self.arg_table = arg_table
        self._subcommand_table = {}
        self._related_items = []
        self.renderer = get_renderer()
        self.doc = ReSTDocument(target='man')

    @property
    def event_class(self):
        """
        Return the ``event_class`` for this object.

        The ``event_class`` is used by the documentation pipeline
        when generating documentation events.  For the event below::

            doc-title.<event_class>.<name>

        The document pipeline would use this property to determine
        the ``event_class`` value.
        """
        pass

    @property
    def name(self):
        """
        Return the name of the wrapped object.

        This would be called by the document pipeline to determine
        the ``name`` to be inserted into the event, as shown above.
        """
        pass

    @property
    def subcommand_table(self):
        """These are the commands that may follow after the help command"""
        return self._subcommand_table

    @property
    def related_items(self):
        """This is list of items that are related to the help command"""
        return self._related_items

    def __call__(self, args, parsed_globals):
        if args:
            subcommand_parser = ArgTableArgParser({}, self.subcommand_table)
            parsed, remaining = subcommand_parser.parse_known_args(args)
            if getattr(parsed, 'subcommand', None) is not None:
                return self.subcommand_table[parsed.subcommand](
                    remaining, parsed_globals
                )

        # Create an event handler for a Provider Document
        instance = self.EventHandlerClass(self)
        # Now generate all of the events for a Provider document.
        # We pass ourselves along so that we can, in turn, get passed
        # to all event handlers.
        docevents.generate_events(self.session, self)
        self.renderer.render(self.doc.getvalue())
        instance.unregister()


class ProviderHelpCommand(HelpCommand):
    """Implements top level help command.

    This is what is called when ``aws help`` is run.

    """

    EventHandlerClass = ProviderDocumentEventHandler

    def __init__(
        self, session, command_table, arg_table, description, synopsis, usage
    ):
        HelpCommand.__init__(self, session, None, command_table, arg_table)
        self.description = description
        self.synopsis = synopsis
        self.help_usage = usage
        self._subcommand_table = None
        self._topic_tag_db = None
        self._related_items = ['aws help topics']

    @property
    def event_class(self):
        return 'aws'

    @property
    def name(self):
        return 'aws'

    @property
    def subcommand_table(self):
        if self._subcommand_table is None:
            if self._topic_tag_db is None:
                self._topic_tag_db = TopicTagDB()
            self._topic_tag_db.load_json_index()
            self._subcommand_table = self._create_subcommand_table()
        return self._subcommand_table

    def _create_subcommand_table(self):
        subcommand_table = {}
        # Add the ``aws help topics`` command to the ``topic_table``
        topic_lister_command = TopicListerCommand(self.session)
        subcommand_table['topics'] = topic_lister_command
        topic_names = self._topic_tag_db.get_all_topic_names()

        # Add all of the possible topics to the ``topic_table``
        for topic_name in topic_names:
            topic_help_command = TopicHelpCommand(self.session, topic_name)
            subcommand_table[topic_name] = topic_help_command
        return subcommand_table


class ServiceHelpCommand(HelpCommand):
    """Implements service level help.

    This is the object invoked whenever a service command
    help is implemented, e.g. ``aws ec2 help``.

    """

    EventHandlerClass = ServiceDocumentEventHandler

    def __init__(
        self, session, obj, command_table, arg_table, name, event_class
    ):
        super().__init__(
            session, obj, command_table, arg_table
        )
        self._name = name
        self._event_class = event_class

    @property
    def event_class(self):
        return self._event_class

    @property
    def name(self):
        return self._name


class OperationHelpCommand(HelpCommand):
    """Implements operation level help.

    This is the object invoked whenever help for a service is requested,
    e.g. ``aws ec2 describe-instances help``.

    """

    EventHandlerClass = OperationDocumentEventHandler

    def __init__(self, session, operation_model, arg_table, name, event_class):
        HelpCommand.__init__(self, session, operation_model, None, arg_table)
        self.param_shorthand = ParamShorthandParser()
        self._name = name
        self._event_class = event_class

    @property
    def event_class(self):
        return self._event_class

    @property
    def name(self):
        return self._name


class TopicListerCommand(HelpCommand):
    EventHandlerClass = TopicListerDocumentEventHandler

    def __init__(self, session):
        super().__init__(session, None, {}, {})

    @property
    def event_class(self):
        return 'topics'

    @property
    def name(self):
        return 'topics'


class TopicHelpCommand(HelpCommand):
    EventHandlerClass = TopicDocumentEventHandler

    def __init__(self, session, topic_name):
        super().__init__(session, None, {}, {})
        self._topic_name = topic_name

    @property
    def event_class(self):
        return 'topics.' + self.name

    @property
    def name(self):
        return self._topic_name



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import sys

from awscli.customizations.codedeploy.systems import DEFAULT_CONFIG_FILE
from awscli.customizations.codedeploy.utils import (
    IAM_USER_ARN_ARG,
    INSTANCE_NAME_ARG,
    validate_iam_user_arn,
    validate_instance_name,
    validate_region,
    validate_tags,
)
from awscli.customizations.commands import BasicCommand
from awscli.utils import create_nested_client


class Register(BasicCommand):
    NAME = 'register'

    DESCRIPTION = (
        "Creates an IAM user for the on-premises instance, if not provided, "
        "and saves the user's credentials to an on-premises instance "
        "configuration file; registers the on-premises instance with AWS "
        "CodeDeploy; and optionally adds tags to the on-premises instance."
    )

    TAGS_SCHEMA = {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "Key": {
                    "description": "The tag key.",
                    "type": "string",
                    "required": True,
                },
                "Value": {
                    "description": "The tag value.",
                    "type": "string",
                    "required": True,
                },
            },
        },
    }

    ARG_TABLE = [
        INSTANCE_NAME_ARG,
        {
            'name': 'tags',
            'synopsis': '--tags <value>',
            'required': False,
            'nargs': '+',
            'schema': TAGS_SCHEMA,
            'help_text': (
                'Optional. The list of key/value pairs to tag the on-premises '
                'instance.'
            ),
        },
        IAM_USER_ARN_ARG,
    ]

    def _run_main(self, parsed_args, parsed_globals):
        params = parsed_args
        params.session = self._session
        validate_region(params, parsed_globals)
        validate_instance_name(params)
        validate_tags(params)
        validate_iam_user_arn(params)

        self.codedeploy = create_nested_client(
            self._session,
            'codedeploy',
            region_name=params.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )
        self.iam = create_nested_client(
            self._session, 'iam', region_name=params.region
        )

        try:
            if not params.iam_user_arn:
                self._create_iam_user(params)
                self._create_access_key(params)
                self._create_user_policy(params)
                self._create_config(params)
            self._register_instance(params)
            if params.tags:
                self._add_tags(params)
            sys.stdout.write(
                f'Copy the on-premises configuration file named {DEFAULT_CONFIG_FILE} to the '
                'on-premises instance, and run the following command on the '
                'on-premises instance to install and configure the AWS '
                'CodeDeploy Agent:\n'
                f'aws deploy install --config-file {DEFAULT_CONFIG_FILE}\n'
            )
        except Exception as e:
            sys.stdout.flush()
            sys.stderr.write(
                'ERROR\n'
                f'{e}\n'
                'Register the on-premises instance by following the '
                'instructions in "Configure Existing On-Premises Instances by '
                'Using AWS CodeDeploy" in the AWS CodeDeploy User '
                'Guide.\n'
            )

    def _create_iam_user(self, params):
        sys.stdout.write('Creating the IAM user... ')
        params.user_name = params.instance_name
        response = self.iam.create_user(
            Path='/AWS/CodeDeploy/', UserName=params.user_name
        )
        params.iam_user_arn = response['User']['Arn']
        sys.stdout.write('DONE\n' f'IamUserArn: {params.iam_user_arn}\n')

    def _create_access_key(self, params):
        sys.stdout.write('Creating the IAM user access key... ')
        response = self.iam.create_access_key(UserName=params.user_name)
        params.access_key_id = response['AccessKey']['AccessKeyId']
        params.secret_access_key = response['AccessKey']['SecretAccessKey']
        sys.stdout.write(
            'DONE\n'
            f'AccessKeyId: {params.access_key_id}\n'
            f'SecretAccessKey: {params.secret_access_key}\n'
        )

    def _create_user_policy(self, params):
        sys.stdout.write('Creating the IAM user policy... ')
        params.policy_name = 'codedeploy-agent'
        params.policy_document = (
            '{\n'
            '    "Version": "2012-10-17",\n'
            '    "Statement": [ {\n'
            '        "Action": [ "s3:Get*", "s3:List*" ],\n'
            '        "Effect": "Allow",\n'
            '        "Resource": "*"\n'
            '    } ]\n'
            '}'
        )
        self.iam.put_user_policy(
            UserName=params.user_name,
            PolicyName=params.policy_name,
            PolicyDocument=params.policy_document,
        )
        sys.stdout.write(
            'DONE\n'
            f'PolicyName: {params.policy_name}\n'
            f'PolicyDocument: {params.policy_document}\n'
        )

    def _create_config(self, params):
        sys.stdout.write(
            f'Creating the on-premises instance configuration file named {DEFAULT_CONFIG_FILE}'
            '...'
        )
        try:
            fd = os.open(
                DEFAULT_CONFIG_FILE,
                os.O_WRONLY | os.O_CREAT | os.O_TRUNC,
                0o600,
            )
            with os.fdopen(fd, 'w') as f:
                os.chmod(DEFAULT_CONFIG_FILE, 0o600)
                f.write(
                    '---\n'
                    f'region: {params.region}\n'
                    f'iam_user_arn: {params.iam_user_arn}\n'
                    f'aws_access_key_id: {params.access_key_id}\n'
                    f'aws_secret_access_key: {params.secret_access_key}\n'
                )
        except OSError as e:
            raise RuntimeError(
                f'Failed to create config file {DEFAULT_CONFIG_FILE}: {e}'
            )
        sys.stdout.write('DONE\n')

    def _register_instance(self, params):
        sys.stdout.write('Registering the on-premises instance... ')
        self.codedeploy.register_on_premises_instance(
            instanceName=params.instance_name, iamUserArn=params.iam_user_arn
        )
        sys.stdout.write('DONE\n')

    def _add_tags(self, params):
        sys.stdout.write('Adding tags to the on-premises instance... ')
        self.codedeploy.add_tags_to_on_premises_instances(
            tags=params.tags, instanceNames=[params.instance_name]
        )
        sys.stdout.write('DONE\n')



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.formatter import FullyBufferedFormatter


class ListRunsFormatter(FullyBufferedFormatter):
    TITLE_ROW_FORMAT_STRING = "       %-50.50s  %-19.19s  %-23.23s"
    FIRST_ROW_FORMAT_STRING = "%4d.  %-50.50s  %-19.19s  %-23.23s"
    SECOND_ROW_FORMAT_STRING = "       %-50.50s  %-19.19s  %-19.19s"

    def _format_response(self, command_name, response, stream):
        self._print_headers(stream)
        for i, obj in enumerate(response):
            self._print_row(i, obj, stream)

    def _print_headers(self, stream):
        stream.write(self.TITLE_ROW_FORMAT_STRING % (
            "Name", "Scheduled Start", "Status"))
        stream.write('\n')
        second_row = (self.SECOND_ROW_FORMAT_STRING % (
            "ID", "Started", "Ended"))
        stream.write(second_row)
        stream.write('\n')
        stream.write('-' * len(second_row))
        stream.write('\n')

    def _print_row(self, index, obj, stream):
        logical_name = obj['@componentParent']
        object_id = obj['@id']
        scheduled_start_date = obj.get('@scheduledStartTime', '')
        status = obj.get('@status', '')
        start_date = obj.get('@actualStartTime', '')
        end_date = obj.get('@actualEndTime', '')
        first_row = self.FIRST_ROW_FORMAT_STRING % (
            index + 1, logical_name, scheduled_start_date, status)
        second_row = self.SECOND_ROW_FORMAT_STRING % (
            object_id, start_date, end_date)
        stream.write(first_row)
        stream.write('\n')
        stream.write(second_row)
        stream.write('\n\n')



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
Remove deprecated commands
--------------------------

This customization removes commands that are either deprecated or not
yet fully supported.

"""
import logging
from functools import partial

LOG = logging.getLogger(__name__)


def register_removals(event_handler):
    cmd_remover = CommandRemover(event_handler)
    cmd_remover.remove(on_event='building-command-table.ses',
                       remove_commands=['delete-verified-email-address',
                                        'list-verified-email-addresses',
                                        'verify-email-address'])
    cmd_remover.remove(on_event='building-command-table.ec2',
                       remove_commands=['import-instance', 'import-volume'])
    cmd_remover.remove(on_event='building-command-table.emr',
                       remove_commands=['run-job-flow', 'describe-job-flows',
                                        'add-job-flow-steps',
                                        'terminate-job-flows',
                                        'list-bootstrap-actions',
                                        'list-instance-groups',
                                        'set-termination-protection',
                                        'set-keep-job-flow-alive-when-no-steps',
                                        'set-visible-to-all-users',
                                        'set-unhealthy-node-replacement'])
    cmd_remover.remove(on_event='building-command-table.kinesis',
                       remove_commands=['subscribe-to-shard'])
    cmd_remover.remove(on_event='building-command-table.lexv2-runtime',
                         remove_commands=['start-conversation'])
    cmd_remover.remove(on_event='building-command-table.lambda',
                         remove_commands=['invoke-with-response-stream'])
    cmd_remover.remove(on_event='building-command-table.sagemaker-runtime',
                         remove_commands=['invoke-endpoint-with-response-stream'])
    cmd_remover.remove(on_event='building-command-table.bedrock-runtime',
                         remove_commands=['invoke-model-with-response-stream',
                                          'invoke-model-with-bidirectional-stream',
                                          'converse-stream'])
    cmd_remover.remove(on_event='building-command-table.bedrock-agent-runtime',
                         remove_commands=['invoke-agent',
                                          'invoke-flow',
                                          'invoke-inline-agent',
                                          'optimize-prompt',
                                          'retrieve-and-generate-stream'])
    cmd_remover.remove(on_event='building-command-table.bedrock-agentcore',
                        remove_commands=['invoke-code-interpreter',
                                         'invoke-agent-runtime-command'])
    cmd_remover.remove(on_event='building-command-table.qbusiness',
                        remove_commands=['chat'])
    cmd_remover.remove(on_event='building-command-table.iotsitewise',
                        remove_commands=['invoke-assistant'])
    cmd_remover.remove(on_event='building-command-table.logs',
                        remove_commands=['get-log-object'])
    cmd_remover.remove(
        on_event='building-command-table.connecthealth',
        remove_commands=['start-medical-scribe-listening-session'],
    )
    cmd_remover.remove(on_event='building-command-table.polly',
                        remove_commands=['start-speech-synthesis-stream'])
    cmd_remover.remove(
        on_event='building-command-table.devops-agent',
        remove_commands=['send-message'],
    )


class CommandRemover(object):
    def __init__(self, events):
        self._events = events

    def remove(self, on_event, remove_commands):
        self._events.register(on_event,
                              self._create_remover(remove_commands))

    def _create_remover(self, commands_to_remove):
        return partial(_remove_commands, commands_to_remove=commands_to_remove)


def _remove_commands(command_table, commands_to_remove, **kwargs):
    # Hooked up to building-command-table.<service>
    for command in commands_to_remove:
        try:
            LOG.debug("Removing operation: %s", command)
            del command_table[command]
        except KeyError:
            LOG.warning("Attempting to delete command that does not exist: %s",
                        command)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module enables the preview-mode customization.

If a service is marked as being in preview mode, then any attempts
to call operations on that service will print a message pointing
the user to alternate solutions.  A user can still access this
service by enabling the service in their config file via:

    [preview]
    servicename=true

or by running:

    aws configure set preview.servicename true

Also any service that is marked as being in preview will *not*
be listed in the help docs, unless the service has been enabled
in the config file as shown above.

"""
import logging
import sys
import textwrap


logger = logging.getLogger(__name__)


PREVIEW_SERVICES = [
    'sdb',
]


def register_preview_commands(events):
    events.register('building-command-table.main', mark_as_preview)


def mark_as_preview(command_table, session, **kwargs):
    # These are services that are marked as preview but are
    # explicitly enabled in the config file.
    allowed_services = _get_allowed_services(session)
    for preview_service in PREVIEW_SERVICES:
        is_enabled = False
        if preview_service in allowed_services:
            # Then we don't need to swap it as a preview
            # service, the user has specifically asked to
            # enable this service.
            logger.debug("Preview service enabled through config file: %s",
                         preview_service)
            is_enabled = True
        original_command = command_table[preview_service]
        preview_cls = type(
            'PreviewCommand',
            (PreviewModeCommandMixin, original_command.__class__), {})
        command_table[preview_service] = preview_cls(
            cli_name=original_command.name,
            session=session,
            service_name=original_command.service_model.service_name,
            is_enabled=is_enabled)
        # We also want to register a handler that will update the
        # description in the docs to say that this is a preview service.
        session.get_component('event_emitter').register_last(
            'doc-description.%s' % preview_service,
            update_description_with_preview)


def update_description_with_preview(help_command, **kwargs):
    style = help_command.doc.style
    style.start_note()
    style.bold(PreviewModeCommandMixin.HELP_SNIPPET.strip())
    # bcdoc does not currently allow for what I'd like to do
    # which is have a code block like:
    #
    # ::
    #    [preview]
    #    service=true
    #
    #    aws configure set preview.service true
    #
    # So for now we're just going to add the configure command
    # to enable this.
    style.doc.write("You can enable this service by running: ")
    # The service name will always be the first element in the
    # event class for the help object
    service_name = help_command.event_class.split('.')[0]
    style.code("aws configure set preview.%s true" % service_name)
    style.end_note()


def _get_allowed_services(session):
    # For a service to be marked as preview, it must be in the
    # [preview] section and it must have a value of 'true'
    # (case insensitive).
    allowed = []
    preview_services = session.full_config.get('preview', {})
    for preview, value in preview_services.items():
        if value == 'true':
            allowed.append(preview)
    return allowed


class PreviewModeCommandMixin(object):
    ENABLE_DOCS = textwrap.dedent("""\
    However, if you'd like to use the "aws {service}" commands with the
    AWS CLI, you can enable this service by adding the following to your CLI
    config file:

        [preview]
        {service}=true

    or by running:

        aws configure set preview.{service} true

    """)
    HELP_SNIPPET = ("AWS CLI support for this service is only "
                    "available in a preview stage.\n")

    def __init__(self, *args, **kwargs):
        self._is_enabled = kwargs.pop('is_enabled')
        super(PreviewModeCommandMixin, self).__init__(*args, **kwargs)

    def __call__(self, args, parsed_globals):
        if self._is_enabled or self._is_help_command(args):
            return super(PreviewModeCommandMixin, self).__call__(
                args, parsed_globals)
        else:
            return self._display_opt_in_message()

    def _is_help_command(self, args):
        return args and args[-1] == 'help'

    def _display_opt_in_message(self):
        sys.stderr.write(self.HELP_SNIPPET)
        sys.stderr.write("\n")
        # Then let them know how to enable this service.
        sys.stderr.write(self.ENABLE_DOCS.format(service=self._service_name))
        return 1



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from awscli.arguments import CustomArgument
from awscli.customizations.emr import helptext
from awscli.customizations.emr import emrutils


def modify_tags_argument(argument_table, **kwargs):
    argument_table['tags'] = TagsArgument('tags', required=True,
                                          help_text=helptext.TAGS, nargs='+')


class TagsArgument(CustomArgument):
    def add_to_params(self, parameters, value):
        if value is None:
            return
        parameters['Tags'] = emrutils.parse_tags(value)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# Declare all the constants used by Lifecycle in this file

# Lifecycle role names
TRUST_POLICY_STATEMENT_FORMAT = '{ \
    "Effect": "Allow", \
    "Principal": { \
        "Federated": "arn:%(AWS_PARTITION)s:iam::%(AWS_ACCOUNT_ID)s:oidc-provider/' \
                                '%(OIDC_PROVIDER)s" \
    }, \
    "Action": "sts:AssumeRoleWithWebIdentity", \
    "Condition": { \
        "StringLike": { \
            "%(OIDC_PROVIDER)s:sub": "system:serviceaccount:%(NAMESPACE)s' \
                                ':emr-containers-sa-*-*-%(AWS_ACCOUNT_ID)s-' \
                                '%(BASE36_ENCODED_ROLE_NAME)s" \
        } \
    } \
}'

TRUST_POLICY_STATEMENT_ALREADY_EXISTS = "Trust policy statement already " \
                                        "exists for role %s. No changes " \
                                        "were made!"

TRUST_POLICY_UPDATE_SUCCESSFUL = "Successfully updated trust policy of role %s"



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os

from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.emr.constants import EC2_ROLE_NAME
from awscli.customizations.emr.constants import EMR_ROLE_NAME

LOG = logging.getLogger(__name__)


def get_configs(session):
    return session.get_scoped_config().get('emr', {})


def get_current_profile_name(session):
    profile_name = session.get_config_variable('profile')
    return 'default' if profile_name is None else profile_name


def get_current_profile_var_name(session):
    return _get_profile_str(session, '.')


def _get_profile_str(session, separator):
    profile_name = session.get_config_variable('profile')
    return 'default' if profile_name is None \
        else 'profile%c%s' % (separator, profile_name)


def is_any_role_configured(session):
    parsed_configs = get_configs(session)
    return True if ('instance_profile' in parsed_configs or
                    'service_role' in parsed_configs) \
        else False


def update_roles(session):
    if is_any_role_configured(session):
        LOG.debug("At least one of the roles is already associated with "
                  "your current profile ")
    else:
        config_writer = ConfigWriter(session)
        config_writer.update_config('service_role', EMR_ROLE_NAME)
        config_writer.update_config('instance_profile', EC2_ROLE_NAME)
        LOG.debug("Associated default roles with your current profile")


class ConfigWriter(object):

    def __init__(self, session):
        self.session = session
        self.section = _get_profile_str(session, ' ')
        self.config_file_writer = ConfigFileWriter()

    def update_config(self, key, value):
        config_filename = \
            os.path.expanduser(self.session.get_config_variable('config_file'))
        updated_config = {'__section__': self.section,
                          'emr': {key: value}}
        self.config_file_writer.update_config(updated_config, config_filename)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from awscli.customizations.s3.syncstrategy.base import BaseSync


LOG = logging.getLogger(__name__)


DELETE = {'name': 'delete', 'action': 'store_true',
          'help_text': (
              "Files that exist in the destination but not in the source are "
              "deleted during sync. Note that files excluded by filters are "
              "excluded from deletion.")}


class DeleteSync(BaseSync):

    ARGUMENT = DELETE

    def determine_should_sync(self, src_file, dest_file):
        dest_file.operation_name = 'delete'
        LOG.debug("syncing: (None) -> %s (remove), file does not "
                  "exist at source (%s) and delete mode enabled",
                  dest_file.src, dest_file.dest)
        return True



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging

from awscli.arguments import CustomArgument

LOG = logging.getLogger(__name__)

# Nested argument member separator
SEP = '.'


class FlattenedArgument(CustomArgument):
    """
    A custom argument which has been flattened from an existing structure. When
    added to the call params it is hydrated back into the structure.

    Supports both an object and a list of objects, in which case the flattened
    parameters will hydrate a list with a single object in it.
    """
    def __init__(self, name, container, prop, help_text='', required=None,
                 type=None, hydrate=None, hydrate_value=None):
        self.type = type
        self._container = container
        self._property = prop
        self._hydrate = hydrate
        self._hydrate_value = hydrate_value
        super(FlattenedArgument, self).__init__(name=name, help_text=help_text,
                                                required=required)

    @property
    def cli_type_name(self):
        return self.type

    def add_to_params(self, parameters, value):
        """
        Hydrate the original structure with the value of this flattened
        argument.

        TODO: This does not hydrate nested structures (``XmlName1.XmlName2``)!
              To do this for now you must provide your own ``hydrate`` method.
        """
        container = self._container.argument_model.name
        cli_type = self._container.cli_type_name
        key = self._property

        LOG.debug('Hydrating {0}[{1}]'.format(container, key))

        if value is not None:
            # Convert type if possible
            if self.type == 'boolean':
                value = not value.lower() == 'false'
            elif self.type in ['integer', 'long']:
                value = int(value)
            elif self.type in ['float', 'double']:
                value = float(value)

            if self._hydrate:
                self._hydrate(parameters, container, cli_type, key, value)
            else:
                if container not in parameters:
                    if cli_type == 'list':
                        parameters[container] = [{}]
                    else:
                        parameters[container] = {}

                if self._hydrate_value:
                    value = self._hydrate_value(value)

                if cli_type == 'list':
                    parameters[container][0][key] = value
                else:
                    parameters[container][key] = value


class FlattenArguments(object):
    """
    Flatten arguments for one or more commands for a particular service from
    a given configuration which maps service call parameters to flattened
    names. Takes in a configuration dict of the form::

        {
            "command-cli-name": {
                "argument-cli-name": {
                    "keep": False,
                    "flatten": {
                        "XmlName": {
                            "name": "flattened-cli-name",
                            "type": "Optional custom type",
                            "required": "Optional custom required",
                            "help_text": "Optional custom docs",
                            "hydrate_value": Optional function to hydrate value,
                            "hydrate": Optional function to hydrate
                        },
                        ...
                    }
                },
                ...
            },
            ...
        }

    The ``type``, ``required`` and ``help_text`` arguments are entirely
    optional and by default are pulled from the model. You should only set them
    if you wish to override the default values in the model.

    The ``keep`` argument determines whether the original command is still
    accessible vs. whether it is removed. It defaults to ``False`` if not
    present, which removes the original argument.

    The keys inside of ``flatten`` (e.g. ``XmlName`` above) can include nested
    references to structures via a colon. For example, ``XmlName1:XmlName2``
    for the following structure::

        {
            "XmlName1": {
                "XmlName2": ...
            }
        }

    The ``hydrate_value`` function takes in a value and should return a value.
    It is only called when the value is not ``None``. Example::

        "hydrate_value": lambda (value): value.upper()

    The ``hydrate`` function takes in a list of existing parameters, the name
    of the container, its type, the name of the container key and its set
    value. For the example above, the container would be
    ``'argument-cli-name'``, the key would be ``'XmlName'`` and the value
    whatever the user passed in. Example::

        def my_hydrate(params, container, cli_type, key, value):
            if container not in params:
                params[container] = {'default': 'values'}

            params[container][key] = value

    It's possible for ``cli_type`` to be ``list``, in which case you should
    ensure that a list of one or more objects is hydrated rather than a
    single object.
    """
    def __init__(self, service_name, configs):
        self.configs = configs
        self.service_name = service_name

    def register(self, cli):
        """
        Register with a CLI instance, listening for events that build the
        argument table for operations in the configuration dict.
        """
        # Flatten each configured operation when they are built
        service = self.service_name
        for operation in self.configs:
            cli.register('building-argument-table.{0}.{1}'.format(service,
                                                                  operation),
                         self.flatten_args)

    def flatten_args(self, command, argument_table, **kwargs):
        # For each argument with a bag of parameters
        for name, argument in self.configs[command.name].items():
            argument_from_table = argument_table[name]
            overwritten = False

            LOG.debug('Flattening {0} argument {1} into {2}'.format(
                command.name, name,
                ', '.join([v['name'] for k, v in argument['flatten'].items()])
            ))

            # For each parameter to flatten out
            for sub_argument, new_config in argument['flatten'].items():
                config = new_config.copy()
                config['container'] = argument_from_table
                config['prop'] = sub_argument

                # Handle nested arguments
                _arg = self._find_nested_arg(
                    argument_from_table.argument_model, sub_argument
                )

                # Pull out docs and required attribute
                self._merge_member_config(_arg, sub_argument, config)

                # Create and set the new flattened argument
                new_arg = FlattenedArgument(**config)
                argument_table[new_config['name']] = new_arg

                if name == new_config['name']:
                    overwritten = True

            # Delete the original argument?
            if not overwritten and ('keep' not in argument or
                                    not argument['keep']):
                del argument_table[name]

    def _find_nested_arg(self, argument, name):
        """
        Find and return a nested argument, if it exists. If no nested argument
        is requested then the original argument is returned. If the nested
        argument cannot be found, then a ValueError is raised.
        """
        if SEP in name:
            # Find the actual nested argument to pull out
            LOG.debug('Finding nested argument in {0}'.format(name))
            for piece in name.split(SEP)[:-1]:
                for member_name, member in argument.members.items():
                    if member_name == piece:
                        argument = member
                        break
                else:
                    raise ValueError('Invalid piece {0}'.format(piece))

        return argument

    def _merge_member_config(self, argument, name, config):
        """
        Merges an existing config taken from the configuration dict with an
        existing member of an existing argument object. This pulls in
        attributes like ``required`` and ``help_text`` if they have not been
        overridden in the configuration dict. Modifies the config in-place.
        """
        # Pull out docs and required attribute
        for member_name, member in argument.members.items():
            if member_name == name.split(SEP)[-1]:
                if 'help_text' not in config:
                    config['help_text'] = member.documentation

                if 'required' not in config:
                    config['required'] = member_name in argument.required_members

                if 'type' not in config:
                    config['type'] = member.type_name

                break



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os

from botocore.model import ServiceModel

from awscli.customizations.commands import BasicCommand


def _get_endpoint_prefix_to_name_mappings(session):
    # Get the mappings of endpoint prefixes to service names from the
    # available service models.
    prefixes_to_services = {}
    for service_name in session.get_available_services():
        service_model = session.get_service_model(service_name)
        prefixes_to_services[service_model.endpoint_prefix] = service_name
    return prefixes_to_services


def _get_service_name(session, endpoint_prefix):
    if endpoint_prefix in session.get_available_services():
        # Check if the endpoint prefix is a pre-existing service.
        # If it is, use that endpoint prefix as the service name.
        return endpoint_prefix
    else:
        # The service may have a different endpoint prefix than its name
        # So we need to determine what the correct mapping may be.

        # Figure out the mappings of endpoint prefix to service names.
        name_mappings = _get_endpoint_prefix_to_name_mappings(session)
        # Determine the service name from the mapping.
        # If it does not exist in the mapping, return the original endpoint
        # prefix.
        return name_mappings.get(endpoint_prefix, endpoint_prefix)


def get_model_location(session, service_definition, service_name=None):
    """Gets the path of where a service-2.json file should go in ~/.aws/models

    :type session: botocore.session.Session
    :param session: A session object

    :type service_definition: dict
    :param service_definition: The json loaded service definition

    :type service_name: str
    :param service_name: The service name to use. If this not provided,
        this will be determined from a combination of available services
        and the service definition.

    :returns: The path to where are model should be placed based on
        the service definition and the current services in botocore.
    """
    # Add the ServiceModel abstraction over the service json definition to
    # make it easier to work with.
    service_model = ServiceModel(service_definition)

    # Determine the service_name if not provided
    if service_name is None:
        endpoint_prefix = service_model.endpoint_prefix
        service_name = _get_service_name(session, endpoint_prefix)
    api_version = service_model.api_version

    # For the model location we only want the custom data path (~/.aws/models
    # not the one set by AWS_DATA_PATH)
    data_path = session.get_component('data_loader').CUSTOMER_DATA_PATH
    # Use the version of the model to determine the file's naming convention.
    service_model_name = (
        'service-%d.json' % int(
            float(service_definition.get('version', '2.0'))))
    return os.path.join(data_path, service_name, api_version,
        service_model_name)


class AddModelCommand(BasicCommand):
    NAME = 'add-model'
    DESCRIPTION = (
        'Adds a service JSON model to the appropriate location in '
        '~/.aws/models. Once the model gets added, CLI commands and Boto3 '
        'clients will be immediately available for the service JSON model '
        'provided.'
    )
    ARG_TABLE = [
        {'name': 'service-model', 'required': True, 'help_text': (
            'The contents of the service JSON model.')},
        {'name': 'service-name', 'help_text': (
            'Overrides the default name used by the service JSON '
            'model to generate CLI service commands and Boto3 clients.')}
    ]

    def _run_main(self, parsed_args, parsed_globals):
        service_definition = json.loads(parsed_args.service_model)

        # Get the path to where the model should be written
        model_location = get_model_location(
            self._session, service_definition, parsed_args.service_name
        )

        # If the service_name/api_version directories do not exist,
        # then create them.
        model_directory = os.path.dirname(model_location)
        if not os.path.exists(model_directory):
            os.makedirs(model_directory)

        # Write the model to the specified location
        with open(model_location, 'wb') as f:
            f.write(parsed_args.service_model.encode('utf-8'))

        return 0



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from awscli.customizations.emr import argumentschema
from awscli.customizations.emr import emrutils
from awscli.customizations.emr import helptext
from awscli.customizations.emr import instancegroupsutils
from awscli.customizations.emr.command import Command


class AddInstanceGroups(Command):
    NAME = 'add-instance-groups'
    DESCRIPTION = 'Adds an instance group to a running cluster.'
    ARG_TABLE = [
        {'name': 'cluster-id', 'required': True,
         'help_text': helptext.CLUSTER_ID},
        {'name': 'instance-groups', 'required': True,
         'help_text': helptext.INSTANCE_GROUPS,
         'schema': argumentschema.INSTANCE_GROUPS_SCHEMA}
    ]

    def _run_main_command(self, parsed_args, parsed_globals):
        parameters = {'JobFlowId': parsed_args.cluster_id}
        parameters['InstanceGroups'] = \
            instancegroupsutils.build_instance_groups(
            parsed_args.instance_groups)

        add_instance_groups_response = emrutils.call(
            self._session, 'add_instance_groups', parameters,
            self.region, parsed_globals.endpoint_url,
            parsed_globals.verify_ssl)

        constructed_result = self._construct_result(
            add_instance_groups_response)

        emrutils.display_response(self._session, 'add_instance_groups',
                                  constructed_result, parsed_globals)
        return 0

    def _construct_result(self, add_instance_groups_result):
        jobFlowId = None
        instanceGroupIds = None
        clusterArn = None
        if add_instance_groups_result is not None:
            jobFlowId = add_instance_groups_result.get('JobFlowId')
            instanceGroupIds = add_instance_groups_result.get(
                'InstanceGroupIds')
            clusterArn = add_instance_groups_result.get('ClusterArn')

        if jobFlowId is not None and instanceGroupIds is not None:
            return {'ClusterId': jobFlowId,
                    'InstanceGroupIds': instanceGroupIds,
                    'ClusterArn': clusterArn}
        else:
            return {}



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


SSL_COMMON_NAMES = {
    "sqs": {
        "af-south-1": "af-south-1.queue.amazonaws.com",
        "ap-east-1": "ap-east-1.queue.amazonaws.com",
        "ap-northeast-1": "ap-northeast-1.queue.amazonaws.com",
        "ap-northeast-2": "ap-northeast-2.queue.amazonaws.com",
        "ap-northeast-3": "ap-northeast-3.queue.amazonaws.com",
        "ap-south-1": "ap-south-1.queue.amazonaws.com",
        "ap-southeast-1": "ap-southeast-1.queue.amazonaws.com",
        "ap-southeast-2": "ap-southeast-2.queue.amazonaws.com",
        "ap-southeast-3": "ap-southeast-3.queue.amazonaws.com",
        "ca-central-1": "ca-central-1.queue.amazonaws.com",
        "eu-central-1": "eu-central-1.queue.amazonaws.com",
        "eu-north-1": "eu-north-1.queue.amazonaws.com",
        "eu-south-1": "eu-south-1.queue.amazonaws.com",
        "eu-west-1": "eu-west-1.queue.amazonaws.com",
        "eu-west-2": "eu-west-2.queue.amazonaws.com",
        "eu-west-3": "eu-west-3.queue.amazonaws.com",
        "me-south-1": "me-south-1.queue.amazonaws.com",
        "sa-east-1": "sa-east-1.queue.amazonaws.com",
        "us-east-1": "queue.amazonaws.com",
        "us-east-2": "us-east-2.queue.amazonaws.com",
        "us-west-1": "us-west-1.queue.amazonaws.com",
        "us-west-2": "us-west-2.queue.amazonaws.com",
        "cn-north-1": "cn-north-1.queue.amazonaws.com.cn",
        "cn-northwest-1": "cn-northwest-1.queue.amazonaws.com.cn",
        "us-gov-west-1": "us-gov-west-1.queue.amazonaws.com",
        "us-isob-east-1": "us-isob-east-1.queue.sc2s.sgov.gov",
    },
    "emr": {
        "af-south-1": "af-south-1.elasticmapreduce.amazonaws.com",
        "ap-east-1": "ap-east-1.elasticmapreduce.amazonaws.com",
        "ap-northeast-1": "ap-northeast-1.elasticmapreduce.amazonaws.com",
        "ap-northeast-2": "ap-northeast-2.elasticmapreduce.amazonaws.com",
        "ap-northeast-3": "ap-northeast-3.elasticmapreduce.amazonaws.com",
        "ap-south-1": "ap-south-1.elasticmapreduce.amazonaws.com",
        "ap-southeast-1": "ap-southeast-1.elasticmapreduce.amazonaws.com",
        "ap-southeast-2": "ap-southeast-2.elasticmapreduce.amazonaws.com",
        "ap-southeast-3": "ap-southeast-3.elasticmapreduce.amazonaws.com",
        "ca-central-1": "ca-central-1.elasticmapreduce.amazonaws.com",
        "eu-north-1": "eu-north-1.elasticmapreduce.amazonaws.com",
        "eu-south-1": "eu-south-1.elasticmapreduce.amazonaws.com",
        "eu-west-1": "eu-west-1.elasticmapreduce.amazonaws.com",
        "eu-west-2": "eu-west-2.elasticmapreduce.amazonaws.com",
        "eu-west-3": "eu-west-3.elasticmapreduce.amazonaws.com",
        "me-south-1": "me-south-1.elasticmapreduce.amazonaws.com",
        "sa-east-1": "sa-east-1.elasticmapreduce.amazonaws.com",
        "us-east-2": "us-east-2.elasticmapreduce.amazonaws.com",
        "us-west-1": "us-west-1.elasticmapreduce.amazonaws.com",
        "us-west-2": "us-west-2.elasticmapreduce.amazonaws.com",
    },
    "rds": {
        "us-east-1": "rds.amazonaws.com",
    },
    "docdb": {
        "us-east-1": "rds.amazonaws.com",
    },
    "neptune": {
        "us-east-1": "rds.amazonaws.com",
    },
    "health": {
        "aws-global": "health.us-east-1.amazonaws.com",
        "af-south-1": "health.us-east-1.amazonaws.com",
        "ap-east-1": "health.us-east-1.amazonaws.com",
        "ap-northeast-1": "health.us-east-1.amazonaws.com",
        "ap-northeast-2": "health.us-east-1.amazonaws.com",
        "ap-northeast-3": "health.us-east-1.amazonaws.com",
        "ap-south-1": "health.us-east-1.amazonaws.com",
        "ap-southeast-1": "health.us-east-1.amazonaws.com",
        "ap-southeast-2": "health.us-east-1.amazonaws.com",
        "ap-southeast-3": "health.us-east-1.amazonaws.com",
        "ca-central-1": "health.us-east-1.amazonaws.com",
        "eu-central-1": "health.us-east-1.amazonaws.com",
        "eu-north-1": "health.us-east-1.amazonaws.com",
        "eu-south-1": "health.us-east-1.amazonaws.com",
        "eu-west-1": "health.us-east-1.amazonaws.com",
        "eu-west-2": "health.us-east-1.amazonaws.com",
        "eu-west-3": "health.us-east-1.amazonaws.com",
        "me-south-1": "health.us-east-1.amazonaws.com",
        "sa-east-1": "health.us-east-1.amazonaws.com",
        "us-east-1": "health.us-east-1.amazonaws.com",
        "us-east-2": "health.us-east-1.amazonaws.com",
        "us-west-1": "health.us-east-1.amazonaws.com",
        "us-west-2": "health.us-east-1.amazonaws.com",
        "cn-north-1": "health.cn-northwest-1.amazonaws.com.cn",
        "cn-northwest-1": "health.cn-northwest-1.amazonaws.com.cn",
        "aws-cn-global": "health.cn-northwest-1.amazonaws.com.cn",
    },
}

REGION_TO_PARTITION_OVERRIDE = {
    "aws-global": "aws",
    "aws-cn-global": "aws-cn",
}


def register_override_ssl_common_name(cli):
    cli.register_last(
        "before-building-argument-table-parser", update_endpoint_url
    )


def update_endpoint_url(session, parsed_globals, **kwargs):
    service = parsed_globals.command
    endpoints = SSL_COMMON_NAMES.get(service)
    # only change url if user has not overridden already themselves
    if endpoints is not None and parsed_globals.endpoint_url is None:
        region = session.get_config_variable("region")
        endpoint_url = endpoints.get(region)
        if endpoint_url is not None:
            parsed_globals.endpoint_url = f"https://{endpoint_url}"
            if service == "health":
                _override_health_region(region, session, parsed_globals)


def _override_health_region(region, session, parsed_globals):
    if region in REGION_TO_PARTITION_OVERRIDE:
        partition = REGION_TO_PARTITION_OVERRIDE[region]
    else:
        partition = session.get_partition_for_region(region)
    if partition == "aws-cn":
        parsed_globals.region = "cn-northwest-1"
    else:
        parsed_globals.region = "us-east-1"
